import torch.nn as nn
import torch
import os
import numpy as np
from vilmedic.networks.blocks.rnn.textencoder import TextEncoder
from vilmedic.networks.blocks.rnn.decoder import ConditionalDecoder
from vilmedic.networks.blocks.rnn.visualdecoder import VisualConditionalDecoder
from vilmedic.networks.blocks.vision import *
from vilmedic.networks.blocks.rnn.evaluation import beam_search
from vilmedic.networks.models.utils import get_n_params


class SumRNN(nn.Module):
    def __init__(self, encoder, decoder, cnn=None, **kwargs):
        super().__init__()
        self.kwargs = kwargs

        encoder_func = encoder.proto
        decoder_func = decoder.proto

        self.enc = eval(encoder_func)(**encoder)
        self.dec = eval(decoder_func)(**decoder, encoder_size=self.enc.ctx_size)

        # Evaluation
        self.eval_func = beam_search

        self.reset_parameters()

        self.cnn = None
        if cnn is not None:
            cnn_func = cnn.pop('proto')
            self.visual_projection = nn.Linear(cnn.pop("visual_embedding_dim"), self.dec.hidden_size)
            self.cnn = eval(cnn_func)(**cnn)

    def reset_parameters(self):
        for name, param in self.named_parameters():
            # Skip 1-d biases and scalars
            if param.requires_grad and param.dim() > 1:
                nn.init.kaiming_normal_(param.data)

        if 'src_emb' in self.kwargs:
            self.set_embeddings(self.kwargs['src_emb'], self.enc.emb)
        if 'tgt_emb' in self.kwargs:
            self.set_embeddings(self.kwargs['tgt_emb'], self.dec.emb)

        if hasattr(self, 'enc') and hasattr(self.enc, 'emb'):
            # Reset padding embedding to 0
            with torch.no_grad():
                self.enc.emb.weight.data[0].fill_(0)

    def encode(self, input_ids, feats=None, images=None, **kwargs):
        # RNN model is batch_first = False
        input_ids = input_ids.permute(1, 0)

        if feats is not None:
            feats = feats.cuda().permute(1, 0, 2)  # RNN takes (n, bs, feat)
            return {'enc': self.enc(input_ids, feats), 'feats': (feats, None)}
        elif images is not None:
            with torch.no_grad():
                feats = self.cnn(images.cuda())
            feats = self.visual_projection(feats)
            feats = feats.permute(1, 0, 2)  # RNN takes (n, bs, feat)
            return {'enc': self.enc(input_ids), 'feats': (feats, None)}
        else:
            return {'enc': self.enc(input_ids)}

    def decode(self, enc_outputs, decoder_input_ids):
        # RNN model is batch_first = False
        decoder_input_ids = decoder_input_ids.permute(1, 0)
        result = self.dec(enc_outputs, decoder_input_ids)
        result['n_items'] = torch.nonzero(decoder_input_ids[1:]).shape[0]
        return result

    def forward(self, input_ids, decoder_input_ids, **kwargs):
        input_ids = input_ids.cuda()
        decoder_input_ids = decoder_input_ids.cuda()

        enc_outputs = self.encode(input_ids, **kwargs)
        result = self.decode(enc_outputs, decoder_input_ids)

        result['loss'] = result['loss'] / result['n_items']

        return result

    def set_embeddings(self, path, obj):
        filename = os.path.basename(path)
        embs = np.load(path, allow_pickle=True)
        assert len(embs) == obj.weight.size(0)
        success = 0
        with torch.no_grad():
            for i, emb in enumerate(embs):
                if emb is None:
                    continue
                obj.weight.data[i] = torch.from_numpy(emb)
                success += 1
        print(filename, success, '/', len(embs), 'words loaded')

    def __repr__(self):
        s = super().__repr__() + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch.nn as nn
from ..rnn.utils import get_n_params, set_embeddings
import copy

from .VQABert.VisualBert import VisualBERTBase
from .VQABert.beam import beam_search
from ..vision.cnn import CNN
import torch

class VQA2SeqHug(nn.Module):
    """
    If proto is mentioned in encoder and decoder dict, loads pretrained models from proto strings.
    Otherwise, loads a BertGenerationEncoder/BertGenerationDecoder model from encoder and decoder dict.
    """

    def __init__(self, cnn, encoder, **kwargs):
        super().__init__()

        # Encoder
        self.enc = VisualBERTBase(**encoder)
        self.backbone = CNN(**cnn).eval()
        # self.visual_projection = nn.Linear(encoder.pop("visual_embedding_dim"), self.enc.config.hidden_size)
        self.eval_func = beam_search
        self.generate_config = {}

    def forward(self, input_ids, images, attention_mask, labels, **kwargs):
        features, _ = self.backbone(images.cuda())
        # features = features.detach()
        out = self.enc(input_ids=input_ids.cuda(),
                       input_mask=attention_mask.cuda(),
                       attention_mask=attention_mask.cuda(),
                       visual_embeddings=features,
                       labels=labels.cuda(),
                       )
        return out
    #
    def train(self, mode: bool = True):
        self.training = mode
        for module in self.children():
            module.train(mode)
        if self.backbone.network.freeze:
            self.backbone.train(False)
        return self

    def __repr__(self):
        # s = super().__repr__() + '\n'
        s = str(type(self.enc).__name__) + '(' + str(self.enc.config) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch.nn as nn
from ..rnn.utils import get_n_params, set_embeddings
import copy

# v4.3.2
from transformers import EncoderDecoderModel
from transformers import AutoModelForCausalLM
from transformers import AutoConfig, EncoderDecoderConfig
from .ImageBert.ImageFeaturesBert import ImageFeaturesBert
from .ImageBert.beam import beam_search
from ..vision.cnn import CNN


class Visual2SeqHug(nn.Module):
    """
    If proto is mentioned in encoder and decoder dict, loads pretrained models from proto strings.
    Otherwise, loads a BertGenerationEncoder/BertGenerationDecoder model from encoder and decoder dict.
    """

    def __init__(self, cnn, encoder, decoder, **kwargs):
        super().__init__()

        # Encoder
        enc = ImageFeaturesBert()
        self.backbone = CNN(**cnn)
        self.visual_projection = nn.Linear(encoder.pop("visual_embedding_dim"), enc.config.hidden_size)

        # Decoder
        decoder_pretrained_model_name_or_path = decoder.pop('proto')
        decoder_config = AutoConfig.from_pretrained(decoder_pretrained_model_name_or_path)
        if decoder_config.is_decoder is False:
            print(
                f"Initializing {decoder_pretrained_model_name_or_path} "
                f"as a decoder model. Cross attention layers are added to "
                f"{decoder_pretrained_model_name_or_path} "
                f"and randomly initialized if "
                f"{decoder_pretrained_model_name_or_path}'s architecture allows for cross attention layers.")
            decoder_config.is_decoder = True
            decoder_config.add_cross_attention = True

        kwargs_decoder = {"config": decoder_config}

        dec = AutoModelForCausalLM.from_pretrained(decoder_pretrained_model_name_or_path, **kwargs_decoder)
        config = EncoderDecoderConfig.from_encoder_decoder_configs(enc.config, dec.config)
        self.enc_dec = EncoderDecoderModel(encoder=enc, decoder=dec, config=config)

        self.enc = self.enc_dec.encoder
        self.dec = self.enc_dec.decoder

        # Tokens
        self.bos_token_id = self.dec.config.bos_token_id
        self.eos_token_id = self.dec.config.eos_token_id
        self.pad_token_id = self.dec.config.pad_token_id

        # Evaluation
        self.eval_func = beam_search
        self.generate_config = {}

        # Embeddings
        if 'src_emb' in kwargs:
            set_embeddings(kwargs['src_emb'], self.enc.embeddings.word_embeddings)
        if 'tgt_emb' in kwargs:
            set_embeddings(kwargs['tgt_emb'], self.dec.bert.embeddings.word_embeddings)

    def forward(self, input_ids, decoder_input_ids, decoder_attention_mask):
        input_ids = input_ids.cuda()
        input_ids, _ = self.backbone(input_ids)
        input_ids = input_ids.detach()
        input_ids = self.visual_projection(input_ids)

        decoder_input_ids = decoder_input_ids.cuda()
        decoder_attention_mask = decoder_attention_mask.cuda()
        out = self.enc_dec(input_ids=input_ids,
                           attention_mask=None,
                           decoder_input_ids=decoder_input_ids,
                           decoder_attention_mask=decoder_attention_mask,
                           labels=decoder_input_ids)
        out = vars(out)
        return out

    def train(self, mode: bool = True):
        self.training = mode
        for module in self.children():
            module.train(mode)
        self.backbone.train(False)
        return self


    def __repr__(self):
        # s = super().__repr__() + '\n'
        s = str(type(self.enc_dec.encoder).__name__) + '(' + str(self.enc_dec.encoder.config) + ')\n'
        s += str(type(self.enc_dec.decoder).__name__) + '(' + str(self.enc_dec.decoder.config) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



# from . import seq2seq
from .beam_helpers import generate
import tqdm
import torch
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions

"""
Performs beam-search and ensembling.
Use this function for huggingface-based model
Either you are using a seq2seq pipeline, in which case your seq2seq attribute should be called "enc_dec".
Or you are using only a decoder, in which case your dec attribute should be called "dec"
"""


def beam_search(models, config, dl):
    # Get huggingface object responsible of generation (encoder_decoder model or decoder model)
    model = models[0]
    if hasattr(model, "enc_dec"):
        hugg_models = [model.enc_dec for model in models]
    elif hasattr(model, "dec"):
        hugg_models = [model.dec for model in models]
    else:
        raise NotImplementedError("Model must have an enc_dec or dec attribute")

    # Get tokenizer and reference sentences from dataloader
    dataset_str = type(dl.dataset).__name__.lower()
    if 'seq2seq' in dataset_str:
        ref_str = 'decoder_input_ids'
        tokenizer = dl.dataset.tgt_tokenizer
        max_len = dl.dataset.tgt_len
    elif 'seq' in dataset_str:
        ref_str = 'input_ids'
        tokenizer = dl.dataset.tokenizer
        max_len = dl.dataset.max_len
    else:
        raise NotImplementedError("Wrong dataset for this evaluation")

    ref_list = []
    hyp_list = []

    with torch.no_grad():
        for batch in tqdm.tqdm(dl):
            batch = {k: v.cuda() for k, v in batch.items()}

            # If huggingface is only the decoder, run the encoder before generate function.
            if hugg_models[0].config.is_decoder:
                encoder_outputs = [{'encoder_outputs': BaseModelOutputWithPoolingAndCrossAttentions(
                    last_hidden_state=model.encoder(**batch))} for model in models]
                # also change batch["attention_mask"] (that is supposed to decoder) to the mask of encoder (which are just ones)
                encoder_batch_size, encoder_sequence_length, _ = encoder_outputs[0][
                    "encoder_outputs"].last_hidden_state.size()
                encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)
                encoder_attention_mask = torch.ones(encoder_hidden_shape).cuda()
                batch["attention_mask"] = encoder_attention_mask
            else:
                encoder_outputs = None

            hyps = generate(hugg_models,
                            batch=batch,
                            encoder_outputs=encoder_outputs,
                            num_return_sequences=1,
                            max_length=max_len,
                            num_beams=config.beam_width,
                            length_penalty=config.length_penalty,
                            bos_token_id=model.bos_token_id,
                            eos_token_id=model.eos_token_id,
                            pad_token_id=model.pad_token_id,
                            )
            # hyps = hugg_models[0].generate(input_ids=batch["input_ids"], attention_mask=batch["attention_mask"],
            #                                num_return_sequences=1,
            #                                max_length=max_len,
            #                                num_beams=config.beam_width,
            #                                length_penalty=config.length_penalty,
            #                                bos_token_id=model.bos_token_id,
            #                                eos_token_id=model.eos_token_id,
            #                                pad_token_id=model.pad_token_id,
            #                                )

            refs = batch[ref_str]
            for h, r in zip(hyps, refs):
                hyp_list.append(tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
                ref_list.append(tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))

        return {'refs': ref_list, 'hyps': hyp_list}



import torch.nn as nn

# v4.3.2
from transformers.models.auto import AutoModel, AutoConfig
from transformers.models.bert_generation import BertGenerationEncoder, BertGenerationConfig

from transformers import RobertaConfig, RobertaModel
from transformers.models.bert.modeling_bert import BertPooler


class EncoderModel(nn.Module):
    """
    If proto is mentioned in encoder dict, loads pretrained models from proto strings.
    Otherwise, loads a BertGenerationEncoder model from encoder dict.
    """

    def __init__(self, encoder, **kwargs):
        super().__init__()
        if encoder.proto is not None:
            path = encoder.pop('proto')
            enc_config = AutoConfig.from_pretrained(path)
            self.encoder = AutoModel.from_pretrained(path, config=enc_config)
        else:
            enc_config = BertGenerationConfig(**encoder)
            self.encoder = BertGenerationEncoder(enc_config)

        if encoder.add_pooling_layer:  # 4 info: roberta already has a pooler layer
            self.pooler = BertPooler(encoder)

    def forward(self, input_ids=None,
                attention_mask=None,
                position_ids=None,
                head_mask=None,
                inputs_embeds=None,
                encoder_hidden_states=None,
                encoder_attention_mask=None,
                past_key_values=None,
                use_cache=None,
                output_attentions=None,
                output_hidden_states=None,
                **kwargs):

        out = self.encoder(input_ids=input_ids,
                           attention_mask=attention_mask,
                           position_ids=position_ids,
                           head_mask=head_mask,
                           inputs_embeds=inputs_embeds,
                           encoder_hidden_states=encoder_hidden_states,
                           encoder_attention_mask=encoder_attention_mask,
                           past_key_values=past_key_values,
                           use_cache=use_cache,
                           output_attentions=output_attentions,
                           output_hidden_states=output_hidden_states,
                           return_dict=True,
                           )

        if hasattr(self, "pooler"):
            pooled_output = self.pooler(hidden_states=out.last_hidden_state)
            setattr(out, "pooler_output", pooled_output)

        out = vars(out)
        return out

    def __repr__(self):
        s = str(type(self.encoder).__name__) + '(' + str(self.encoder.config) + ')\n'
        return s



import gdown
import os
import zipfile
from pathlib import Path
from tqdm import tqdm
import sys


def download_images(data_name, file_id, unzip_dir):
    zip_name = os.path.join("data", data_name) + ".zip"
    target_dir = os.path.join(unzip_dir, data_name)

    if not os.path.exists(target_dir):
        gdown.download(url="https://drive.google.com/uc?id=" + file_id,
                       output=zip_name,
                       quiet=False)
        print("Unzipping...")
        with zipfile.ZipFile(zip_name, 'r') as zf:
            for member in tqdm(zf.infolist(), desc='Extracting '):
                zf.extract(member, unzip_dir)

        os.remove(zip_name)
        return
    print('{} already exists'.format(target_dir))


def download_model(file_id, unzip_dir):
    if not os.path.exists(unzip_dir):
        os.makedirs(unzip_dir, exist_ok=True)
        gdown.download_folder(id=file_id,
                              output=unzip_dir,
                              quiet=False)


def download_data(file_id, unzip_dir):
    if not os.path.exists(unzip_dir):
        os.makedirs(unzip_dir, exist_ok=True)
        gdown.download_folder(id=file_id,
                              output=unzip_dir,
                              quiet=False)



import json
import os
import torch
import functools
import torch.nn as nn
from tqdm import tqdm
import numpy as np
from torchvision.utils import save_image
from dalle_pytorch import DALLE as PyDALLE
from .VAE import VAE
from vilmedic.networks.models.utils import get_n_params

import torch.nn.functional as F
from einops import rearrange
from .dalle_forward import forward


def evaluation(models, config, dl, **kwargs):
    losses = []
    model = models[0]

    if config.generate_images is not None and config.generate_images:
        pbar = tqdm(dl, total=len(dl))
        for batch in pbar:
            for input_id in batch["input_ids"]:
                print("Inputed text:", dl.dataset.tokenizer.decode(input_id, skip_special_tokens=True))
                print("Generating {} images in dir {}".format(config.num_images, config.ckpt_dir))
                # text_tokens = repeat(input_id.unsqueeze(0), '() n -> b n', b=config.num_images)
                for i in range(config.num_images):
                    print("doing", i)
                    output = model.dalle.generate_images(input_id.unsqueeze(0).cuda(), filter_thres=0.9)
                    print("done output")
                    save_image(output, os.path.join(config.ckpt_dir, '{}.jpg'.format(i)), normalize=True)
                import sys
                sys.exit()

    else:
        pbar = tqdm(dl, total=len(dl))
        for batch in pbar:
            out = model(**batch)
            out['loss'] = out['loss'].mean()
            losses.append(out['loss'].cpu().data.numpy())
        return {'loss': np.ndarray.mean(np.array(losses))}


class DALLE(nn.Module):
    def __init__(self,
                 vae,
                 dalle,
                 forward_batch_size=256,
                 **kwargs):
        super(DALLE, self).__init__()

        assert 'ckpt' in vae, 'please specify vae checkpoint'
        self.vae = VAE(**vae)

        vae_weights = torch.load(vae.pop('ckpt'))["model"]
        vae_weights = {name.replace('module.', ''): param for name, param in vae_weights.items()}
        self.vae.load_state_dict(vae_weights, strict=True)

        self.dalle_config = dalle
        self.dalle = PyDALLE(
            vae=self.vae.vae,  # automatically infer (1) image sequence length and (2) number of image tokens
            **self.dalle_config
        )

        self.fbs = forward_batch_size
        self.dalle.forward = functools.partial(forward, self.dalle)
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, **kwargs):
        images = images.cuda()
        input_ids = input_ids.cuda()

        bs = images.shape[0]
        forward_logits = []
        forward_images = []
        forward_texts = []

        # Forward pass
        for i in range(int(bs / min(self.fbs, bs))):
            input_id = input_ids[i * self.fbs:(i + 1) * self.fbs]
            image = images[i * self.fbs:(i + 1) * self.fbs]
            logits, text, image = self.dalle(input_id, image, return_loss=False)

            forward_logits.append(logits)
            forward_images.append(image)
            forward_texts.append(text)

        forward_logits = torch.cat(forward_logits)
        forward_images = torch.cat(forward_images)
        forward_texts = torch.cat(forward_texts)

        # Compute loss
        offsetted_image = forward_images + self.dalle.num_text_tokens
        labels = torch.cat((forward_texts[:, 1:], offsetted_image), dim=1)
        logits = rearrange(forward_logits, 'b n c -> b c n')
        loss_text = F.cross_entropy(logits[:, :, :self.dalle.text_seq_len], labels[:, :self.dalle.text_seq_len])
        loss_img = F.cross_entropy(logits[:, :, self.dalle.text_seq_len:], labels[:, self.dalle.text_seq_len:])
        loss = (loss_text + self.dalle.loss_img_weight * loss_img) / (self.dalle.loss_img_weight + 1)

        return {'loss': loss}

    def __repr__(self):
        s = str(type(self.dalle).__name__) + '(' + str(
            json.dumps(dict(self.dalle_config), indent=4, sort_keys=True)) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
import itertools
import numpy as np
from tqdm import tqdm


def evaluation(models, config, dl, **kwargs):
    logits = []
    labels = []
    losses = []
    attentions = []

    post_processing = {}

    pbar = tqdm(dl, total=len(dl))
    for batch in pbar:

        label = batch['labels']
        batch_size = len(label)

        batch = {k: v.cuda() for k, v in batch.items()}
        results = [model(**batch) for model in models]
        # iterating over the batch, stacking refs and hyps
        for i in range(batch_size):
            logits.append([r['output'][i].data.cpu().numpy() for r in results])
            labels.append(label[i].data.cpu().numpy())

        # Do we have attention weights?
        if 'attentions' in results[0]:
            # attentions will be of size num_models x bs x num_layers x num_heads x seq_len x seq_len
            for i in range(batch_size):
                attentions.append([torch.stack(list(r['attentions']))[:, i, :].data.cpu().numpy() for r in results])

        # getting average loss
        losses.append(np.mean([r['loss'].cpu().item() for r in results]))

    preds = np.mean(np.array(logits), axis=1)
    loss = np.mean(np.array(losses))

    if attentions:
        post_processing["attentions"] = np.array(attentions)

    return {**{'loss': loss, 'refs': np.array(labels), 'hyps': np.array(preds)}, **post_processing}



import os
import numpy as np
import json
import torch.nn.functional as F
import torch
# import nlgeval

from .NLG import ROUGEScorer
from .NLG import METEORScorer
from .NLG import BLEUScorer
from .NLG import Cider
import mauve
from .NLG import CiderD
# from nlgeval import NLGEval

from .CheXbert.chexbert import CheXbert

from sklearn.metrics import classification_report, roc_auc_score


def compute_scores(metrics, refs, hyps, split, seed, config, epoch):
    scores = dict()
    # If metric is None or empty list
    if metrics is None or not metrics:
        return scores

    assert refs is not None and hyps is not None, \
        "You specified metrics but your evaluation does not return hyps nor refs"

    assert len(refs) == len(hyps), 'refs and hyps must have same length : {} vs {}'.format(len(refs), len(hyps))

    # Dump
    ckpt_dir = config.ckpt_dir
    base = os.path.join(ckpt_dir, '{}_{}_{}'.format(split, seed, '{}'))
    refs_file = base.format('refs.txt')
    hyps_file = base.format('hyps.txt')
    metrics_file = base.format('metrics.txt')

    with open(refs_file, 'w') as f:
        f.write('\n'.join(map(str, refs)))
        f.close()

    with open(hyps_file, 'w') as f:
        f.write('\n'.join(map(str, hyps)))
        f.close()

    for metric in metrics:
        if metric == 'BLEU':
            scores["BLEU"] = round(BLEUScorer().compute(refs_file, hyps_file), 2)
        elif metric == 'ROUGE1':
            scores["ROUGE1"] = round(ROUGEScorer(rouges=['rouge1']).compute(refs, hyps)[0] * 100, 2)
        elif metric == 'ROUGE2':
            scores["ROUGE2"] = round(ROUGEScorer(rouges=['rouge2']).compute(refs, hyps)[0] * 100, 2)
        elif metric == 'ROUGEL':
            scores["ROUGEL"] = round(ROUGEScorer(rouges=['rougeL']).compute(refs, hyps)[0] * 100, 2)
        elif metric == 'METEOR':
            scores["METEOR"] = round(METEORScorer().compute(refs_file, hyps_file) * 100, 2)
        elif metric == 'CIDER':
            scores["CIDER"] = Cider().compute_score(refs, hyps)
        elif metric == 'MAUVE':
            scores["MAUVE"] = mauve.compute_mauve(p_text=refs, q_text=hyps, device_id=0, max_text_length=256,
                                                  featurize_model_name=config.mauve_featurize_model_name or "gpt2-large")
        elif metric == 'accuracy':
            scores["accuracy"] = round(np.mean(np.array(refs) == np.argmax(hyps, axis=-1)) * 100, 2)
        elif metric == 'f1-score':
            scores["f1-score"] = classification_report(refs, np.argmax(hyps, axis=-1))
        elif metric == 'auroc':
            scores["auroc"] = roc_auc_score(refs, F.softmax(torch.from_numpy(hyps), dim=-1).numpy(), multi_class="ovr")
        elif metric == 'chexbert':
            scores["chexbert"], scores["chexbert-5"] = CheXbert(refs_filename=base.format('refs.chexbert.txt'),
                                                                hyps_filename=base.format('hyps.chexbert.txt'))(hyps,
                                                                                                                refs)
            scores["chexbert-5_micro avg_f1-score"] = scores["chexbert-5"]["micro avg"]["f1-score"]
        else:
            raise NotImplementedError(metric)

    with open(metrics_file, 'a+') as f:
        f.write(json.dumps({
            'split': split,
            'epoch': epoch,
            'scores': scores
        }, indent=4, sort_keys=False))
    return scores



import torch
import tqdm
import os
import logging

from vilmedic.constants import EXTRA_CACHE_DIR

import torch.nn as nn
import pandas as pd

from collections import OrderedDict

from transformers import BertTokenizer
from transformers import BertModel, AutoModel, AutoConfig
from sklearn.metrics import classification_report
import numpy as np

logging.getLogger("urllib3").setLevel(logging.ERROR)


def generate_attention_masks(batch, source_lengths):
    """Generate masks for padded batches to avoid self-attention over pad tokens
    @param batch (Tensor): tensor of token indices of shape (batch_size, max_len)
                           where max_len is length of longest sequence in the batch
    @param source_lengths (List[Int]): List of actual lengths for each of the
                           sequences in the batch
    @param device (torch.device): device on which data should be

    @returns masks (Tensor): Tensor of masks of shape (batch_size, max_len)
    """
    masks = torch.ones(batch.size(0), batch.size(1), dtype=torch.float)
    for idx, src_len in enumerate(source_lengths):
        masks[idx, src_len:] = 0
    return masks.cuda()


class bert_labeler(nn.Module):
    def __init__(self, p=0.1, clinical=False, freeze_embeddings=False, pretrain_path=None, inference=False):
        """ Init the labeler module
        @param p (float): p to use for dropout in the linear heads, 0.1 by default is consistant with
                          transformers.BertForSequenceClassification
        @param clinical (boolean): True if Bio_Clinical BERT desired, False otherwise. Ignored if
                                   pretrain_path is not None
        @param freeze_embeddings (boolean): true to freeze bert embeddings during training
        @param pretrain_path (string): path to load checkpoint from
        """
        super(bert_labeler, self).__init__()

        if pretrain_path is not None:
            self.bert = BertModel.from_pretrained(pretrain_path)
        elif clinical:
            self.bert = AutoModel.from_pretrained("emilyalsentzer/Bio_ClinicalBERT")
        elif inference:
            config = AutoConfig.from_pretrained('bert-base-uncased')
            self.bert = AutoModel.from_config(config)
        else:
            self.bert = BertModel.from_pretrained('bert-base-uncased')

        if freeze_embeddings:
            for param in self.bert.embeddings.parameters():
                param.requires_grad = False

        self.dropout = nn.Dropout(p)
        # size of the output of transformer's last layer
        hidden_size = self.bert.pooler.dense.in_features
        # classes: present, absent, unknown, blank for 12 conditions + support devices
        self.linear_heads = nn.ModuleList([nn.Linear(hidden_size, 4, bias=True) for _ in range(13)])
        # classes: yes, no for the 'no finding' observation
        self.linear_heads.append(nn.Linear(hidden_size, 2, bias=True))

    def forward(self, source_padded, attention_mask):
        """ Forward pass of the labeler
        @param source_padded (torch.LongTensor): Tensor of word indices with padding, shape (batch_size, max_len)
        @param attention_mask (torch.Tensor): Mask to avoid attention on padding tokens, shape (batch_size, max_len)
        @returns out (List[torch.Tensor])): A list of size 14 containing tensors. The first 13 have shape
                                            (batch_size, 4) and the last has shape (batch_size, 2)
        """
        # shape (batch_size, max_len, hidden_size)
        final_hidden = self.bert(source_padded, attention_mask=attention_mask)[0]
        # shape (batch_size, hidden_size)
        cls_hidden = final_hidden[:, 0, :].squeeze(dim=1)
        cls_hidden = self.dropout(cls_hidden)
        out = []
        for i in range(14):
            out.append(self.linear_heads[i](cls_hidden))
        return out


def tokenize(impressions, tokenizer):
    imp = impressions.str.strip()
    imp = imp.replace('\n', ' ', regex=True)
    imp = imp.replace('\s+', ' ', regex=True)
    impressions = imp.str.strip()
    new_impressions = []
    for i in (range(impressions.shape[0])):
        tokenized_imp = tokenizer.tokenize(impressions.iloc[i])
        if tokenized_imp:  # not an empty report
            res = tokenizer.encode_plus(tokenized_imp)['input_ids']
            if len(res) > 512:  # length exceeds maximum size
                # print("report length bigger than 512")
                res = res[:511] + [tokenizer.sep_token_id]
            new_impressions.append(res)
        else:  # an empty report
            new_impressions.append([tokenizer.cls_token_id, tokenizer.sep_token_id])
    return new_impressions


class CheXbert(nn.Module):
    def __init__(self, refs_filename=None, hyps_filename=None):
        super(CheXbert, self).__init__()
        self.refs_filename = refs_filename
        self.hyps_filename = hyps_filename

        self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
        self.model = bert_labeler(inference=True)
        state_dict = torch.load(os.path.join(EXTRA_CACHE_DIR, "chexbert.pth"))['model_state_dict']
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            name = k.replace('module.', '')  # remove `module.`
            new_state_dict[name] = v
        # load params
        self.model.load_state_dict(new_state_dict, strict=False)
        self.model = self.model.cuda().eval()

        for name, param in self.model.named_parameters():
            param.requires_grad = False

        self.target_names = [
            "Enlarged Cardiomediastinum", "Cardiomegaly", "Lung Opacity", "Lung Lesion", "Edema",
            "Consolidation", "Pneumonia", "Atelectasis", "Pneumothorax", "Pleural Effusion", "Pleural Other",
            "Fracture", "Support Devices", "No Finding"]

        self.target_names_5 = ["Cardiomegaly", "Edema", "Consolidation", "Atelectasis", "Pleural Effusion"]
        self.target_names_5_index = np.where(np.isin(self.target_names, self.target_names_5))[0]

    def get_label(self, report, mode="rrg"):
        impressions = pd.Series([report])
        out = tokenize(impressions, self.tokenizer)
        batch = torch.LongTensor([o for o in out])
        src_len = [b.shape[0] for b in batch]
        attn_mask = generate_attention_masks(batch, src_len)
        out = self.model(batch.cuda(), attn_mask)
        out = [out[j].argmax(dim=1).item() for j in range(len(out))]
        v = []
        if mode == "rrg":
            for c in out:
                if c == 0:
                    v.append('')
                if c == 3:
                    v.append(1)
                if c == 2:
                    v.append(0)
                if c == 1:
                    v.append(1)
            v = [1 if (isinstance(l, int) and l > 0) else 0 for l in v]

        elif mode == "classification":
            # https://github.com/stanfordmlgroup/CheXbert/blob/master/src/label.py#L124
            for c in out:
                if c == 0:
                    v.append('')
                if c == 3:
                    v.append(-1)
                if c == 2:
                    v.append(0)
                if c == 1:
                    v.append(1)
        else:
            raise NotImplementedError(mode)

        return v

    def forward(self, hyps, refs):
        if self.refs_filename is None:
            refs_chexbert = [self.get_label(l.strip()) for l in refs]
        else:
            if os.path.exists(self.refs_filename):
                refs_chexbert = [eval(l.strip()) for l in open(self.refs_filename).readlines()]
            else:
                refs_chexbert = [self.get_label(l.strip()) for l in refs]
                open(self.refs_filename, 'w').write('\n'.join(map(str, refs_chexbert)))

        hyps_chexbert = [self.get_label(l.strip()) for l in hyps]
        if self.hyps_filename is not None:
            open(self.hyps_filename, 'w').write('\n'.join(map(str, hyps_chexbert)))

        refs_chexbert_5 = [np.array(r)[self.target_names_5_index] for r in refs_chexbert]
        hyps_chexbert_5 = [np.array(h)[self.target_names_5_index] for h in hyps_chexbert]

        cr = classification_report(refs_chexbert, hyps_chexbert, target_names=self.target_names, output_dict=True)
        cr_5 = classification_report(refs_chexbert_5, hyps_chexbert_5, target_names=self.target_names_5,
                                     output_dict=True)

        return cr, cr_5

    def train(self, mode: bool = True):
        mode = False  # force False
        self.training = mode
        for module in self.children():
            module.train(mode)
        return self



import torch
import torch.nn as nn
from torchvision.models import *
from torchxrayvision.models import DenseNet as XrvDenseNet, ResNet as XrvResNet
from .vgg_hgap import *


def get_network(backbone, output_layer, pretrained, weights=None, **kwargs):
    """
    Create sub-network given a backbone and an output_layer
    """
    # Create avgpool for densenet, doesnt exist as such
    if 'densenet' in backbone and output_layer == 'avgpool':
        sub_network = get_network(backbone, 'features', pretrained, **kwargs)
        sub_network.add_module('relu', nn.ReLU(inplace=True))
        sub_network.add_module('avgpool', nn.AdaptiveAvgPool2d((1, 1)))
        sub_network.add_module('flatten', nn.Flatten(1))
        return sub_network

    # torchxrayvision
    if "xrv" in backbone.lower():
        network = eval(backbone)(weights=weights)
        if hasattr(network, 'model'):  # XrvResNet Fix
            network = network.model
    else:
        network = eval(backbone)(pretrained=pretrained, **kwargs)

    if output_layer is not None and (not output_layer == 'classifier'):
        layers = [n for n, _ in network.named_children()]
        assert output_layer in layers, '{} not in {}'.format(output_layer, layers)
        sub_network = []
        for n, c in network.named_children():
            sub_network.append(c)
            if n == output_layer:
                break
        network = nn.Sequential(*sub_network)

    return network


class CNN(nn.Module):
    def __init__(self, backbone, dropout_out, permute, freeze=True, output_layer=None, pretrained=True,
                 visual_embedding_dim=None, **kwargs):
        super(CNN, self).__init__()
        self.backbone = backbone
        self.output_layer = output_layer
        self.permute = permute
        self.freeze = freeze
        self.pretrained = pretrained

        self.cnn = get_network(self.backbone, self.output_layer, self.pretrained, **kwargs)
        self.dropout_out = nn.Dropout(p=dropout_out)

        assert permute in ["batch_first", "spatial_first", "no_permute"]

        if freeze:
            for name, param in self.cnn.named_parameters():
                param.requires_grad = False

    def forward(self, images, **kwargs):
        out = self.cnn(images)
        out = self.dropout_out(out)
        if self.permute == "no_permute":
            out = out
        elif self.permute == "batch_first":
            out = out.view(*out.size()[:2], -1).permute(0, 2, 1)
            if out.shape[1] == 1:  # avgpool case
                out = out.squeeze(1)
        elif self.permute == "spatial_first":
            out = out.view(*out.size()[:2], -1).permute(2, 0, 1)
        else:
            raise NotImplementedError()

        return out

    def train(self, mode: bool = True):
        if self.freeze:
            mode = False
        self.training = mode
        for module in self.children():
            module.train(mode)
        return self

    def __repr__(self):
        s = str(self.backbone) + '(output_layer=' + self.output_layer + ', dropout_out=' + str(
            self.dropout_out.p) + ', freeze=' + str(self.freeze) + ', pretrained=' + str(self.pretrained) + \
            ('\n classifier= {}'.format(self.cnn.classifier) if self.output_layer == 'classifier' else '' + ')')
        return s



from vilmedic.networks.models.mvqa.MVQA import MVQA

from vilmedic.networks.models.rrs.SumRNN import SumRNN
from vilmedic.networks.models.rrs.SumHugMulti import SumHugMulti
from vilmedic.networks.models.rrs.RRS import RRS
from vilmedic.networks.models.rrs.SumHugMono_SCST import SumHugMono_SCST

from vilmedic.networks.models.selfsup.clip.VAE import VAE
from vilmedic.networks.models.selfsup.clip.DALLE import DALLE
from vilmedic.networks.models.selfsup.clip.CLIP import CLIP

from vilmedic.networks.models.rrg.RRG import RRG
from vilmedic.networks.models.rrg.RRG_PPO import RRG_PPO
from vilmedic.networks.models.rrg.RRG_PPO_x import RRG_PPO_x
from vilmedic.networks.models.rrg.RRG_SCST import RRG_SCST

from vilmedic.networks.models.selfsup.conVIRT import ConVIRT
from vilmedic.networks.models.selfsup.SimCLR import SimCLR
from vilmedic.networks.models.selfsup.GLoRIA import GLoRIA


# from vilmedic.networks.models.others.RCNN import RCNN

from vilmedic.networks.models.others.sim_mcan import SIM_MCAN




from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import torch
from torch.nn import functional as F
from transformers.generation_utils import GreedySearchOutput, SampleOutput, BeamSearchOutput, BeamSampleOutput, logging, \
    ModelOutput, BeamSearchEncoderDecoderOutput, BeamSearchDecoderOnlyOutput, BeamSearchScorer

from transformers.generation_beam_search import BeamScorer, BeamHypotheses, UserDict

from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions

from transformers.generation_logits_process import (
    LogitsProcessorList,
)
import copy

logger = logging.get_logger(__name__)


def prepare_inputs_for_generation(
        input_ids, past=None, attention_mask=None, use_cache=None, encoder_outputs=None, **kwargs
):
    if past is not None:
        input_ids = input_ids[:, -1:]

    input_dict = {
        "attention_mask": attention_mask,
        "decoder_attention_mask": input_ids.new_ones(input_ids.shape),
        "decoder_input_ids": input_ids,
        "encoder_outputs": encoder_outputs,
        "past_key_values": past,
        "use_cache": use_cache,
    }
    return input_dict


def beam_search(
        hugg_model,
        hugg_models,
        input_ids: torch.LongTensor,
        beam_scorer: BeamScorer,
        logits_processor: Optional[LogitsProcessorList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        model_kwargs_list=None
) -> Union[BeamSearchOutput, torch.LongTensor]:
    batch_size = len(beam_scorer._beam_hyps)
    num_beams = beam_scorer.num_beams
    batch_beam_size, cur_len = input_ids.shape
    assert (
            num_beams * batch_size == batch_beam_size
    ), "Batch dimension of `input_ids` should be {num_beams * batch_size}, but is {batch_beam_size}."

    beam_scores = torch.zeros((batch_size, num_beams), dtype=torch.float, device=input_ids.device)
    beam_scores[:, 1:] = -1e9
    beam_scores = beam_scores.view((batch_size * num_beams,))

    while cur_len < max_length:
        models_inputs = [prepare_inputs_for_generation(input_ids, **m)
                         for hm, m in zip(hugg_models, model_kwargs_list)]

        outputs = [hm(**model_inputs,
                          return_dict=True,
                          output_attentions=output_attentions,
                          output_hidden_states=output_hidden_states)
                   for hm, model_inputs in zip(hugg_models, models_inputs)]

        next_token_logits = sum(o.logits[:, -1, :] for o in outputs)

        # adjust tokens for Bart, *e.g.*
        next_token_logits = hugg_model.adjust_logits_during_generation(
            next_token_logits, cur_len=cur_len, max_length=max_length
        )

        next_token_scores = F.log_softmax(next_token_logits, dim=-1)  # (batch_size * num_beams, vocab_size)
        next_token_scores = logits_processor(input_ids, next_token_scores)
        next_token_scores = next_token_scores + beam_scores[:, None].expand_as(next_token_scores)

        # reshape for beam search
        vocab_size = next_token_scores.shape[-1]
        next_token_scores = next_token_scores.view(batch_size, num_beams * vocab_size)

        next_token_scores, next_tokens = torch.topk(
            next_token_scores, 2 * num_beams, dim=1, largest=True, sorted=True
        )

        next_indices = next_tokens // vocab_size
        next_tokens = next_tokens % vocab_size

        # stateless
        beam_outputs = beam_scorer.process(
            input_ids,
            next_token_scores,
            next_tokens,
            next_indices,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
        )
        beam_scores = beam_outputs["next_beam_scores"]
        beam_next_tokens = beam_outputs["next_beam_tokens"]
        beam_idx = beam_outputs["next_beam_indices"]

        input_ids = torch.cat([input_ids[beam_idx, :], beam_next_tokens.unsqueeze(-1)], dim=-1)

        cur_len = cur_len + 1

        model_kwargs_list = [
            hm._update_model_kwargs_for_generation(o, m, is_encoder_decoder=True)
            for hm, o, m in zip(hugg_models, outputs, model_kwargs_list)]

        for m in model_kwargs_list:
            if m["past"] is not None:
                m["past"] = hugg_model._reorder_cache(m["past"], beam_idx)

        if beam_scorer.is_done:
            break

    sequence_outputs = beam_scorer.finalize(
        input_ids, beam_scores, next_tokens, next_indices, pad_token_id=pad_token_id, eos_token_id=eos_token_id
    )

    return sequence_outputs["sequences"]


def generate(
        hugg_models,
        batch,
        encoder_outputs=None,
        num_return_sequences=None,
        max_length=None,
        num_beams=None,
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        length_penalty=None,

):
    hugg_model = hugg_models[0]
    assert "attention_mask" in batch
    assert "input_ids" in batch

    device = hugg_model.device
    do_early_stopping = hugg_model.config.early_stopping
    num_beam_groups = hugg_model.config.num_beam_groups
    length_penalty = 1.0 if length_penalty is None else length_penalty

    # The element to tile are the inputs of the decoder
    to_tile = ["last_hidden_state"] if not hasattr(hugg_model, 'to_tile') else hugg_model.to_tile
    assert isinstance(to_tile, list)

    batch = {
        argument: value for argument, value in batch.items() if not argument.startswith("decoder_")
    }

    # input_ids is now decoder_input_ids are just [bos_token_id] * bs
    input_ids = hugg_model._prepare_decoder_input_ids_for_generation(
        batch["input_ids"], bos_token_id=bos_token_id
    )

    # Huggingface seq2seq : Get each encoder output.
    encoder_outputs = [{"encoder_outputs": hm.get_encoder()(**{**batch, **{'return_dict': True}})}
                       for hm in hugg_models]

    # get distribution pre_processing samplers
    logits_processor = hugg_model._get_logits_processor(
        repetition_penalty=None,
        no_repeat_ngram_size=None,
        encoder_no_repeat_ngram_size=None,
        encoder_input_ids=input_ids,
        bad_words_ids=None,
        min_length=None,
        max_length=max_length,
        eos_token_id=eos_token_id,
        prefix_allowed_tokens_fn=None,
        num_beams=num_beams,
        num_beam_groups=num_beam_groups,
        diversity_penalty=None,
        remove_invalid_values=None,
        forced_bos_token_id=None,
        forced_eos_token_id=None,
    )

    if num_return_sequences > num_beams:
        raise ValueError("`num_return_sequences` has to be smaller or equal to `num_beams`.")

    beam_scorer = BeamSearchScorer(
        batch_size=input_ids.shape[0],
        max_length=max_length,
        num_beams=num_beams,
        device=device,
        length_penalty=length_penalty,
        do_early_stopping=do_early_stopping,
        num_beam_hyps_to_keep=num_return_sequences,
    )

    # tile what is needed for decoder
    # input_ids and model_kwargs["encoder_outputs"] to beam_width
    # maybe other arguments

    expanded_return_idx = (
        torch.arange(input_ids.shape[0]).view(-1, 1).repeat(1, num_beams).view(-1).to(input_ids.device)
    )
    input_ids = input_ids.index_select(0, expanded_return_idx)
    for output in encoder_outputs:
        output["attention_mask"] = batch["attention_mask"].index_select(0, expanded_return_idx)
        for t in to_tile:
            to_tile_tensor = getattr(output["encoder_outputs"], t)
            setattr(output["encoder_outputs"], t,
                    to_tile_tensor.index_select(0, expanded_return_idx.to(to_tile_tensor.device)))

    return beam_search(
        hugg_model,
        hugg_models,
        input_ids,
        beam_scorer,
        logits_processor=logits_processor,
        max_length=max_length,
        pad_token_id=pad_token_id,
        eos_token_id=eos_token_id,
        model_kwargs_list=encoder_outputs
    )


# -*- coding: utf-8 -*-
from torch import nn
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence
from .layers.ff import FF
import torch


def sort_batch(seqbatch):
    """Sorts torch tensor of integer indices by decreasing order."""
    # 0 is padding_idx
    omask = (seqbatch != 0).long()
    olens = omask.sum(0)
    slens, sidxs = torch.sort(olens, descending=True)
    oidxs = torch.sort(sidxs)[1]
    return (oidxs, sidxs, slens.data.tolist(), omask.float())


class TextEncoder(nn.Module):
    """A recurrent encoder with embedding layer.

    Arguments:
        input_size (int): Embedding dimensionality.
        hidden_size (int): RNN hidden state dimensionality.
        n_vocab (int): Number of tokens for the embedding layer.
        rnn_type (str): RNN Type, i.e. GRU or LSTM.
        num_layers (int, optional): Number of stacked RNNs (Default: 1).
        bidirectional (bool, optional): If `False`, the RNN is unidirectional.
        dropout_rnn (float, optional): Inter-layer dropout rate only
            applicable if `num_layers > 1`. (Default: 0.)
        dropout_emb(float, optional): Dropout rate for embeddings (Default: 0.)
        dropout_ctx(float, optional): Dropout rate for the
            encodings/annotations (Default: 0.)
        emb_maxnorm(float, optional): If given, renormalizes embeddings so
            that their norm is the given value.
        emb_gradscale(bool, optional): If `True`, scales the gradients
            per embedding w.r.t. to its frequency in the batch.
        proj_dim(int, optional): If not `None`, add a final projection
            layer. Can be used to adapt dimensionality for decoder.
        proj_activ(str, optional): Non-linearity for projection layer.
            `None` or `linear` does not apply any non-linearity.
        layer_norm(bool, optional): Apply layer normalization at the
            output of the encoder.
root
    Input:
        x (Tensor): A tensor of shape (n_timesteps, n_samples)
            including the integer token indices for the given batch.

    Output:
        hs (Tensor): A tensor of shape (n_timesteps, n_samples, hidden)
            that contains encoder hidden states for all timesteps. If
            bidirectional, `hs` is doubled in size in the last dimension
            to contain both directional states.
        mask (Tensor): A binary mask of shape (n_timesteps, n_samples)
            that may further be used in attention and/or decoder. `None`
            is returned if batch contains only sentences with same lengths.
    """

    def __init__(self, input_size, hidden_size, n_vocab, rnn_type,
                 num_layers=1, bidirectional=True, layer_norm=False,
                 enc_init=None, enc_init_activ=None, enc_init_size=None,
                 ctx_mul=None, ctx_mul_activ=None, ctx_mul_size=None,
                 dropout_rnn=0, dropout_emb=0, dropout_ctx=0,
                 proj_dim=None, proj_activ=None, **kwargs):
        super().__init__()

        self.rnn_type = rnn_type.upper()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.n_vocab = n_vocab
        self.num_layers = num_layers
        self.bidirectional = bidirectional
        self.proj_dim = proj_dim
        self.proj_activ = proj_activ
        self.layer_norm = layer_norm

        # For dropout btw layers, only effective if num_layers > 1
        self.dropout_rnn = dropout_rnn

        # Our other custom dropouts after embeddings and annotations
        self.dropout_emb = dropout_emb
        self.dropout_ctx = dropout_ctx

        self.ctx_size = self.hidden_size
        # Doubles its size because of concatenation
        if self.bidirectional:
            self.ctx_size *= 2

        # Embedding dropout
        self.do_emb = nn.Dropout(self.dropout_emb)

        # Create embedding layer
        self.emb = nn.Embedding(self.n_vocab, self.input_size,
                                padding_idx=0)

        # Create fused/cudnn encoder according to the requested type
        RNN = getattr(nn, self.rnn_type)
        self.enc = RNN(self.input_size, self.hidden_size,
                       self.num_layers, bias=True, batch_first=False,
                       dropout=self.dropout_rnn,
                       bidirectional=self.bidirectional)

        output_layers = []
        if self.proj_dim:
            output_layers.append(
                FF(self.ctx_size, self.proj_dim, activ=self.proj_activ))
            self.ctx_size = self.proj_dim
        if self.layer_norm:
            output_layers.append(torch.nn.LayerNorm(self.ctx_size))
        if self.dropout_ctx > 0:
            output_layers.append(nn.Dropout(p=self.dropout_ctx))
        self.output = nn.Sequential(*output_layers)

        # Enc init
        self.enc_init = enc_init
        self.enc_init_activ = enc_init_activ
        self.enc_init_size = enc_init_size

        assert self.enc_init in (None, 'feats'), \
            "dec_init '{}' not known".format(enc_init)

        if self.enc_init is not None:
            self.tile_factor = self.num_layers
            if self.bidirectional:
                self.tile_factor *= 2
            self.ff_enc_init = FF(self.enc_init_size, self.hidden_size, activ=self.enc_init_activ)

        # ctx_mul
        self.ctx_mul = ctx_mul
        self.ctx_mul_activ = ctx_mul_activ
        self.ctx_mul_size = ctx_mul_size
        assert self.ctx_mul in (None, 'feats'), \
            "ctx_mul '{}' not known".format(ctx_mul)

        if self.ctx_mul is not None:
            self.out_dim = self.hidden_size
            if self.bidirectional:
                self.out_dim *= 2
            self.ff_ctx_mul = FF(self.ctx_mul_size, self.out_dim, activ=self.ctx_mul_activ)

    def f_init(self, feats):
        """Returns the initial h_0 for the encoder."""
        if self.enc_init is None:
            return None
        elif self.enc_init == 'feats':
            return self.ff_enc_init(feats).expand(self.tile_factor, -1, -1).contiguous()
        else:
            raise NotImplementedError(self.enc_init)

    def forward(self, x, feats=None, **kwargs):
        h0 = self.f_init(feats)

        # Non-homogeneous batches possible
        # sort the batch by decreasing length of sequences
        # oidxs: to recover original order
        # sidxs: idxs to sort the batch
        # slens: lengths in sorted order for pack_padded_sequence()
        oidxs, sidxs, slens, mask = sort_batch(x)

        # Fetch embeddings for the sorted batch
        embs = self.do_emb(self.emb(x[:, sidxs]))
        # Pack and encode
        packed_emb = pack_padded_sequence(embs, slens)

        # We ignore last_state since we don't use it
        self.enc.flatten_parameters()
        packed_hs, _ = self.enc(packed_emb, h0)

        # Get hidden states and revert the order
        hs = pad_packed_sequence(packed_hs)[0][:, oidxs]
        if self.ctx_mul is not None:
            hs = hs * self.ff_ctx_mul(feats)
        return self.output(hs), mask



import torch.nn as nn
from vilmedic.networks.models.utils import get_n_params
# v4.3.2

from vilmedic.networks.blocks.vision import *
from vilmedic.networks.blocks.huggingface.decoder.decoder_model import DecoderModel
from vilmedic.networks.blocks.huggingface.decoder.evaluation import evaluation


class RRG(nn.Module):

    def __init__(self, decoder, cnn, **kwargs):
        super().__init__()

        # Decoder
        self.dec = DecoderModel(decoder)

        # Encoder
        visual_embedding_dim = cnn.pop("visual_embedding_dim")
        cnn = eval(cnn.pop('proto'))(**cnn)
        visual_projection = nn.Linear(visual_embedding_dim, self.dec.decoder.config.hidden_size)
        self.enc = nn.Sequential(cnn, visual_projection)

        # Evaluation
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, encoder_outputs=None, **kwargs):
        images = images.cuda()
        input_ids = input_ids.cuda()
        attention_mask = attention_mask.cuda()

        if encoder_outputs is None:
            encoder_outputs = self.encode(images)

        out = self.dec(input_ids=input_ids,
                       attention_mask=attention_mask,
                       encoder_outputs=encoder_outputs,
                       **kwargs)

        return out

    # Necessary for generation
    def encode(self, images, **kwargs):
        images = images.cuda()
        return self.enc(images)

    def __repr__(self):
        s = "RRG\n"
        s += str(self.enc) + '\n'
        s += str(self.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch.nn as nn
from vilmedic.networks.models.utils import get_n_params
from torchvision.models.detection.faster_rcnn import FasterRCNN
from torchvision.models.detection.backbone_utils import resnet_fpn_backbone
from .coco.coco_eval import CocoEvaluator
from .coco.coco_utils import get_coco_api_from_dataset
from tqdm import tqdm
import numpy as np



def evaluation(models, config, dl, **kwargs):
    losses = []
    pbar = tqdm(dl, total=len(dl))
    iou_types = ["bbox"]
    coco = get_coco_api_from_dataset(dl.dataset)
    coco_evaluator = CocoEvaluator(coco, iou_types)

    for batch in pbar:
        results = [model(**batch) for model in models]
        results = results[0]
        res = {target["image_id"].item(): output for target, output in zip(batch["targets"], results)}
        coco_evaluator.update(res)
        break

    coco_evaluator.synchronize_between_processes()
    # accumulate predictions from all images
    coco_evaluator.accumulate()
    coco_evaluator.summarize()
    troll
    return {'loss': loss}

class RCNN(nn.Module):

    def __init__(self, **kwargs):
        super().__init__()
        backbone = resnet_fpn_backbone('resnet50', pretrained=False, trainable_layers=5)
        self.model = FasterRCNN(backbone,
                                num_classes=14)
        # Evaluation
        self.eval_func = evaluation
        # self.eval_func = beam_search

    def forward(self, images, targets, **kwargs):
        targets = [{k: v.cuda() for k, v in t.items()} for t in targets]
        outputs = self.model(images.cuda(), targets=targets)

        if not self.training:
            return [{k: v.cpu() for k, v in t.items()} for t in outputs]
        else:
            return {'loss': sum(loss for loss in outputs.values())}

    def __repr__(self):
        s = "RCNN\n"
        s += str(self.model) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



# --------------------------------------------------------
# OpenVQA
# Written by Yuhao Cui https://github.com/cuiyuhao1996
# --------------------------------------------------------

from .fc import FC, MLP

import torch.nn as nn
import torch.nn.functional as F
import torch


# ------------------------------
# ---- Flatten the sequence ----
# ------------------------------

class AttFlat(nn.Module):
    def __init__(self, __C):
        super(AttFlat, self).__init__()
        self.__C = __C

        self.mlp = MLP(
            in_size=__C.HIDDEN_SIZE,
            mid_size=__C.FLAT_MLP_SIZE,
            out_size=__C.FLAT_GLIMPSES,
            dropout_r=__C.DROPOUT_R,
            use_relu=True
        )

        self.linear_merge = nn.Linear(
            __C.HIDDEN_SIZE * __C.FLAT_GLIMPSES,
            __C.FLAT_OUT_SIZE
        )

    def forward(self, x, x_mask):
        att = self.mlp(x)
        att = att.masked_fill(
            x_mask.squeeze(1).squeeze(1).unsqueeze(2),
            -1e9
        )
        att = F.softmax(att, dim=1)

        att_list = []
        for i in range(self.__C.FLAT_GLIMPSES):
            att_list.append(
                torch.sum(att[:, :, i: i + 1] * x, dim=1)
            )

        x_atted = torch.cat(att_list, dim=1)
        x_atted = self.linear_merge(x_atted)

        return x_atted





from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import torch
from torch.nn import functional as F

from transformers.file_utils import ModelOutput
from transformers.generation_beam_search import BeamScorer, BeamSearchScorer
from transformers.generation_logits_process import (
    EncoderNoRepeatNGramLogitsProcessor,
    ForcedBOSTokenLogitsProcessor,
    ForcedEOSTokenLogitsProcessor,
    HammingDiversityLogitsProcessor,
    InfNanRemoveLogitsProcessor,
    LogitsProcessorList,
    MinLengthLogitsProcessor,
    NoBadWordsLogitsProcessor,
    NoRepeatNGramLogitsProcessor,
    PrefixConstrainedLogitsProcessor,
    RepetitionPenaltyLogitsProcessor,
    TemperatureLogitsWarper,
    TopKLogitsWarper,
    TopPLogitsWarper,
)
from transformers.generation_stopping_criteria import (
    MaxLengthCriteria,
    MaxTimeCriteria,
    StoppingCriteriaList,
    validate_stopping_criteria,
)
from transformers.utils import logging
from transformers.generation_utils import BeamSearchOutput
from transformers.generation_utils import BeamSearchEncoderDecoderOutput
from transformers.generation_utils import BeamSearchDecoderOnlyOutput


def prepare_inputs_for_generation(self, input_ids, past=None, attention_mask=None, **model_kwargs):
    input_shape = input_ids.shape
    # if model is used as a decoder in encoder-decoder model, the decoder attention mask is created on the fly
    if attention_mask is None:
        attention_mask = input_ids.new_ones(input_shape)

    # cut decoder_input_ids if past is used
    if past is not None:
        input_ids = input_ids[:, -1:]
    return {"input_ids": input_ids, "attention_mask": attention_mask, "past_key_values": past, **model_kwargs}


def beam_search(
        self,
        input_ids: torch.LongTensor,
        beam_scorer: BeamScorer,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        **model_kwargs,
) -> Union[BeamSearchOutput, torch.LongTensor]:
    # init values
    logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
    stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
    max_length = max_length if max_length is not None else self.config.max_length
    validate_stopping_criteria(stopping_criteria, max_length)
    pad_token_id = pad_token_id if pad_token_id is not None else self.config.pad_token_id
    eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id
    output_scores = output_scores if output_scores is not None else self.config.output_scores
    output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate is not None else self.config.return_dict_in_generate
    )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    decoder_attentions = () if (return_dict_in_generate and output_attentions) else None
    cross_attentions = () if (return_dict_in_generate and output_attentions) else None
    decoder_hidden_states = () if (return_dict_in_generate and output_hidden_states) else None

    # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
    if return_dict_in_generate and self.config.is_encoder_decoder:
        encoder_attentions = model_kwargs["encoder_outputs"].get("attentions") if output_attentions else None
        encoder_hidden_states = (
            model_kwargs["encoder_outputs"].get("hidden_states") if output_hidden_states else None
        )

    batch_size = len(beam_scorer._beam_hyps)
    num_beams = beam_scorer.num_beams

    batch_beam_size, cur_len = input_ids.shape

    assert (
            num_beams * batch_size == batch_beam_size
    ), "Batch dimension of `input_ids` should be {num_beams * batch_size}, but is {batch_beam_size}."

    beam_scores = torch.zeros((batch_size, num_beams), dtype=torch.float, device=input_ids.device)
    beam_scores[:, 1:] = -1e9
    beam_scores = beam_scores.view((batch_size * num_beams,))

    # Added by JB
    hf_models = model_kwargs.pop("hf_models")
    model_kwargs_list = model_kwargs.pop("encoders_hidden_states")
    ###

    while cur_len < max_length:
        # Changed by JB
        models_inputs = [hf_models[0].prepare_inputs_for_generation(input_ids, **m) for m in model_kwargs_list]
        outputs = [hf(**model_inputs,
                      return_dict=True,
                      output_attentions=output_attentions,
                      output_hidden_states=output_hidden_states)
                   for hf, model_inputs in zip(hf_models, models_inputs)]

        next_token_logits = sum(o.logits[:, -1, :] for o in outputs)
        ###

        # hack: adjust tokens for Marian. For Marian we have to make sure that the `pad_token_id`
        # cannot be generated both before and after the `F.log_softmax` operation.
        next_token_logits = self.adjust_logits_during_generation(
            next_token_logits, cur_len=cur_len, max_length=max_length
        )

        next_token_scores = F.log_softmax(next_token_logits, dim=-1)  # (batch_size * num_beams, vocab_size)

        next_token_scores = logits_processor(input_ids, next_token_scores)
        next_token_scores = next_token_scores + beam_scores[:, None].expand_as(next_token_scores)

        # Store scores, attentions and hidden_states when required
        if return_dict_in_generate:
            if output_scores:
                scores += (next_token_scores,)
            if output_attentions:
                decoder_attentions += (
                    (outputs.decoder_attentions,) if self.config.is_encoder_decoder else (outputs.attentions,)
                )
                if self.config.is_encoder_decoder:
                    cross_attentions += (outputs.cross_attentions,)

            if output_hidden_states:
                decoder_hidden_states += (
                    (outputs.decoder_hidden_states,)
                    if self.config.is_encoder_decoder
                    else (outputs.hidden_states,)
                )

        # reshape for beam search
        vocab_size = next_token_scores.shape[-1]
        next_token_scores = next_token_scores.view(batch_size, num_beams * vocab_size)

        next_token_scores, next_tokens = torch.topk(
            next_token_scores, 2 * num_beams, dim=1, largest=True, sorted=True
        )

        next_indices = next_tokens // vocab_size
        next_tokens = next_tokens % vocab_size

        # stateless
        beam_outputs = beam_scorer.process(
            input_ids,
            next_token_scores,
            next_tokens,
            next_indices,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
        )
        beam_scores = beam_outputs["next_beam_scores"]
        beam_next_tokens = beam_outputs["next_beam_tokens"]
        beam_idx = beam_outputs["next_beam_indices"]

        input_ids = torch.cat([input_ids[beam_idx, :], beam_next_tokens.unsqueeze(-1)], dim=-1)

        cur_len = cur_len + 1

        # Changed by JB
        model_kwargs_list = [
            hf._update_model_kwargs_for_generation(o, m, is_encoder_decoder=True)
            for hf, o, m in zip(hf_models, outputs, model_kwargs_list)]

        for m in model_kwargs_list:
            if m["past"] is not None:
                m["past"] = self._reorder_cache(m["past"], beam_idx)
        ###

        if beam_scorer.is_done:
            break

        if stopping_criteria(input_ids, scores):
            break

    sequence_outputs = beam_scorer.finalize(
        input_ids, beam_scores, next_tokens, next_indices, pad_token_id=pad_token_id, eos_token_id=eos_token_id
    )

    if return_dict_in_generate:
        if not output_scores:
            sequence_outputs["sequence_scores"] = None
        if self.config.is_encoder_decoder:
            return BeamSearchEncoderDecoderOutput(
                sequences=sequence_outputs["sequences"],
                sequences_scores=sequence_outputs["sequence_scores"],
                scores=scores,
                encoder_attentions=encoder_attentions,
                encoder_hidden_states=encoder_hidden_states,
                decoder_attentions=decoder_attentions,
                cross_attentions=cross_attentions,
                decoder_hidden_states=decoder_hidden_states,
            )
        else:
            return BeamSearchDecoderOnlyOutput(
                sequences=sequence_outputs["sequences"],
                sequences_scores=sequence_outputs["sequence_scores"],
                scores=scores,
                attentions=decoder_attentions,
                hidden_states=decoder_hidden_states,
            )
    else:
        return sequence_outputs["sequences"]



import os
import sys
import glob
import torch

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from utils import get_args, get, print_args, get_seed
from logger import set_logger
from vilmedic.executors import Validator, create_model


def get_n_best(mode):
    n = 1
    # checking if args is formatted as best-n
    if '-' in mode:
        n = int(mode.split('-')[-1])
    return n


def get_ckpts(path, mode):
    ckpts = glob.glob(path)

    # Sort by score
    ckpts = sorted(ckpts, reverse=True)

    # Getting n-best models
    if 'best' in mode:
        n = get_n_best(mode)
        ckpts = ckpts[:n]
    return ckpts


def main():
    config, override = get_args()
    config.ckpt_dir = os.path.join(config.ckpt_dir, config.name)

    ensemble_config = get(config, 'ensemblor')
    seed = '{}_{}'.format(ensemble_config.mode, get_seed())

    set_logger(config.ckpt_dir, seed)

    # Nice printing the args
    print_args(config, ['ensemblor'], seed, override)

    # Create validator, dont give any models yet
    evaluator = Validator(config=ensemble_config,
                          models=None,
                          seed=seed,
                          from_training=False)

    # fetching all ckpt according to 'mode"
    ckpts = get_ckpts(os.path.join(evaluator.config.ckpt_dir, '*.pth'), ensemble_config.mode)

    # if specific checkpoint is specified
    if ensemble_config.ckpt is not None:
        ckpts = [os.path.join(evaluator.config.ckpt_dir, ensemble_config.ckpt)]

    if not ckpts:
        evaluator.logger.settings("No checkpoints found")
        sys.exit()

    evaluator.logger.settings("Checkpoints are {}".format("\n".join(ckpts)))

    # Give models to evaluator
    evaluator.models = [create_model(config=ensemble_config,
                                     # we give the first evaluation split's dataloader to model
                                     dl=evaluator.splits[0][1],
                                     logger=evaluator.logger,
                                     state_dict=torch.load(ckpt)).cuda().eval() for ckpt in ckpts]

    # Boom
    evaluator.start()


if __name__ == "__main__":
    main()



import os
import json
import torch
import logging
from .utils import create_data_loader, get_eval_func
from vilmedic.scorers.scores import compute_scores
from vilmedic.scorers.post_processing import post_processing


class InitValidator(object):
    def __init__(self, config, models, seed, from_training):
        self.seed = seed
        self.config = config
        self.from_training = from_training

        # Logger
        self.logger = logging.getLogger(str(seed))

        self.models = models
        self.metrics = config.metrics
        self.post_processing = config.post_processing
        self.epoch = 0

        # Evaluation splits
        self.splits = [(split, create_data_loader(self.config, split, self.config.logger, called_by_validator=True))
                       for split in self.config.splits]


class Validator(InitValidator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def start(self):
        assert isinstance(self.models, list)
        self.scores = []
        self.models = [m.eval() for m in self.models]

        for split, dl in self.splits:
            self.logger.info('Running split: {} by ensembling {} models. '
                             'Using {}.'.format(split,
                                                len(self.models),
                                                type(dl.batch_sampler.sampler).__name__,
                                                ))

            eval_func = get_eval_func(self.models)
            with torch.no_grad():
                results = eval_func(models=self.models,
                                    config=self.config,
                                    dl=dl,
                                    from_training=self.from_training)

            # model must return at least loss or (refs and hyps)
            # TODO check refs and hyps together
            assert type(results) is dict and \
                   any(key in results for key in ['loss', 'refs', 'hyps']), \
                self.logger.error('Evaluation func does not return any evaluation keys')

            scores = dict()

            # Handle loss
            scores['loss'] = float(results.pop("loss", 0.0))

            # Handle metrics
            metrics = compute_scores(metrics=self.metrics,
                                     refs=results.pop('refs', None),
                                     hyps=results.pop('hyps', None),
                                     split=split,
                                     seed=self.seed,
                                     config=self.config,
                                     epoch=self.epoch
                                     )
            scores.update(metrics)

            # Dumping things for potential post processing
            post_processing(post_processing=self.post_processing,
                            results=results,
                            split=split,
                            seed=self.seed,
                            ckpt_dir=self.config.ckpt_dir,
                            epoch=self.epoch,
                            dl=dl
                            )

            # Logging scores
            self.logger.info(json.dumps(scores, indent=4, sort_keys=False))

            # Saving the metrics for current split
            self.scores.append(scores)



from PIL import Image
import numpy as np
import cv2

PAPER_EXT = {".gloria_chx": "gloria_chx_open_image"}


def gloria_chx_open_image(img):
    def _resize_img(img, scale):
        """
        Args:
            img - image as numpy array (cv2)
            scale - desired output image-size as scale x scale
        Return:
            image resized to scale x scale with shortest dimension 0-padded
        """
        size = img.shape
        max_dim = max(size)
        max_ind = size.index(max_dim)

        # Resizing
        if max_ind == 0:
            # image is heigher
            wpercent = scale / float(size[0])
            hsize = int((float(size[1]) * float(wpercent)))
            desireable_size = (scale, hsize)
        else:
            # image is wider
            hpercent = scale / float(size[1])
            wsize = int((float(size[0]) * float(hpercent)))
            desireable_size = (wsize, scale)
        resized_img = cv2.resize(
            img, desireable_size[::-1], interpolation=cv2.INTER_AREA
        )  # this flips the desireable_size vector

        # Padding
        if max_ind == 0:
            # height fixed at scale, pad the width
            pad_size = scale - resized_img.shape[1]
            left = int(np.floor(pad_size / 2))
            right = int(np.ceil(pad_size / 2))
            top = int(0)
            bottom = int(0)
        else:
            # width fixed at scale, pad the height
            pad_size = scale - resized_img.shape[0]
            top = int(np.floor(pad_size / 2))
            bottom = int(np.ceil(pad_size / 2))
            left = int(0)
            right = int(0)
        resized_img = np.pad(
            resized_img, [(top, bottom), (left, right)], "constant", constant_values=0
        )

        return resized_img

    x = cv2.imread(str(img), 0)
    x = _resize_img(x, 256)
    img = Image.fromarray(x).convert("RGB")
    return img



import pathlib
from setuptools import setup, find_packages

setup(
    name='vilmedic',
    version='1.0.6',
    description='ViLMedic is a modular framework for multimodal research at the intersection of vision and language in the medical field.',
    author='Jean-Benoit Delbrouck',
    license='MIT',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='medical nlp deep-learning pytorch',
    python_requires='~=3.9',
    install_requires=['Cython', 'appdirs==1.4.4', 'omegaconf==2.0.6', 'torchvision==0.9.1', 'rouge_score',
                      'scikit_image==0.18.2',
                      'scikit-learn==0.24.2', 'pydicom==2.2.0', 'transformers==4.5.1', 'tokenizers==0.10.3',
                      'seaborn==0.11.1', 'gdown==4.2.0',
                      'dalle-pytorch==1.4.2', 'torchxrayvision==0.0.32',
                      'torch==1.8.1', 'pytorch-lightning==1.4.2', 'pytorch-metric-learning==0.9.99',
                      'torch-optimizer==0.1.0', 'umap-learn==0.5.2', 'opencv-python==4.5.4.60',
                      'mauve-text', 'numba==0.54.1', 'torchmetrics==0.5.0', 'numpy==1.20.3'],
    include_package_data=True,
    exclude_package_data={'': ['.git']},
    packages=find_packages(exclude=["bin"]),
    scripts=[str(p) for p in pathlib.Path('bin/scripts').glob('*')],
    zip_safe=False)



# --------------------------------------------------------
# OpenVQA
# Written by Yuhao Cui https://github.com/cuiyuhao1996
# --------------------------------------------------------

from vilmedic.networks.mcan.base_cfgs import BaseCfgs


class Cfgs():
    def __init__(self):
        super(Cfgs, self).__init__()

        self.LAYER = 6
        self.HIDDEN_SIZE = 512
        self.BBOXFEAT_EMB_SIZE = 2048
        self.FF_SIZE = 2048
        self.MULTI_HEAD = 8
        self.DROPOUT_R = 0.1
        self.FLAT_MLP_SIZE = 512
        self.FLAT_GLIMPSES = 1
        self.FLAT_OUT_SIZE = 1024
        self.USE_AUX_FEAT = False
        self.USE_BBOX_FEAT = False
        self.BBOX_NORMALIZE = True



# -*- coding: utf-8 -*-
# Filename: cider.py
#
# Description: Describes the class to compute the CIDEr
#              (Consensus-Based Image Description Evaluation) Metric
# by Vedantam, Zitnick, and Parikh (http://arxiv.org/abs/1411.5726)
#
# Creation Date: Sun Feb  8 14:16:54 2015
#
# Authors: Ramakrishna Vedantam <vrama91@vt.edu> and
# Tsung-Yi Lin <tl483@cornell.edu>

from .cider_scorer import CiderScorer


class Cider:
    """Main Class to compute the CIDEr metric."""

    def __init__(self, test=None, refs=None, n=4, sigma=6.0):
        # set cider to sum over 1 to 4-grams
        self._n = n
        # set the standard deviation parameter for gaussian penalty
        self._sigma = sigma

    def compute_score(self, gts, res):
        """Main function to compute CIDEr score

        Arguments:
            hypo_for_image (dict): dictionary with key <image> and
                value <tokenized hypothesis / candidate sentence>
            ref_for_image (dict): dictionary with key <image> and value
                <tokenized reference sentence>

        Returns:
            cider (float): computed CIDEr score for the corpus
        """

        cider_scorer = CiderScorer(n=self._n, sigma=self._sigma)
        res = {i: [v] for i, v in enumerate(res)}
        gts = {i: [v] for i, v in enumerate(gts)}
        for id in sorted(gts.keys()):
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert isinstance(hypo, list)
            assert isinstance(ref, list)
            assert len(hypo) == 1
            assert len(ref) > 0

            cider_scorer += (hypo[0], ref)

        (score, scores) = cider_scorer.compute_score()

        return score

    def method(self):
        return "CIDEr"



# from . import seq2seq
# from .beam_helpers import generate
from ..beam_helpers import generate
import tqdm
import torch
import os
import collections

def beam_search(models, config, dl):
    ref_list = []
    hyp_list = collections.defaultdict(list)

    with torch.no_grad():
        for batch in tqdm.tqdm(dl):
            ref_list.extend(batch['labels'].data.cpu().numpy())
            for i, m in enumerate(models):
                logits = m(**batch)["logits"]
                hyp_list[i].extend(logits)

        hyp_list = torch.stack([torch.stack(v) for v in hyp_list.values()])
        hyp_list = hyp_list.mean(0)
        hyp_list = torch.argmax(hyp_list, dim=-1).data.cpu().numpy()

        answers = [dl.dataset.labels_dict.idx2label(h) for h in hyp_list]

        f = open(os.path.join(dl.dataset.ckpt_dir, dl.dataset.split + '_answers'), 'w')
        f.write('\n'.join(answers))
        return 0.0, ref_list, hyp_list



import tqdm
import torch
import functools
from .beam_search import beam_search


def evaluation(models, config, dl, **kwargs):
    hf_models = [model.dec.decoder for model in models]

    # We are in an ensembling scenario, we override huggingface beam-search function
    hf_models[0].beam_search = functools.partial(beam_search, hf_models[0])

    # Get tokenizer and reference sentences from dataloader
    ref_str = 'input_ids'
    tokenizer = dl.dataset.tokenizer
    max_len = dl.dataset.tokenizer_max_len
    ref_list = []
    hyp_list = []

    with torch.no_grad():
        for batch in tqdm.tqdm(dl):
            batch = {k: v.cuda() for k, v in batch.items()}
            batch_size = batch['images'].shape[0]
            expanded_return_idx = (
                torch.arange(batch_size).view(-1, 1).repeat(1, config.beam_width).view(-1).cuda()
            )

            model_kwargs = {
                "encoders_hidden_states":
                    [{"encoder_hidden_states": hf.encode(**batch).index_select(0, expanded_return_idx)} for hf in
                     models],
                "hf_models": hf_models
            }

            hyps = hf_models[0].generate(
                input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * hf_models[0].config.bos_token_id,
                num_return_sequences=1,
                max_length=max_len,
                num_beams=config.beam_width,
                length_penalty=config.length_penalty,
                bos_token_id=hf_models[0].config.bos_token_id,
                eos_token_id=hf_models[0].config.eos_token_id,
                pad_token_id=hf_models[0].config.pad_token_id,
                use_cache=True,
                **model_kwargs
            )

            refs = batch[ref_str]
            for h, r in zip(hyps, refs):
                hyp_list.append(tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
                ref_list.append(tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))

        return {'refs': ref_list, 'hyps': hyp_list}



import re
import six
from nltk.stem import porter
from nltk.tokenize import wordpunct_tokenize

# https://github.com/cuhksz-nlp/R2Gen/blob/main/modules/tokenizers.py
# Generating Radiology Reports via Memory-driven Transformer
def r2gen_clean_report(report):
    report_cleaner = lambda t: t.replace('\n', ' ').replace('__', '_').replace('__', '_').replace('__', '_') \
        .replace('__', '_').replace('__', '_').replace('__', '_').replace('__', '_').replace('  ', ' ') \
        .replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ') \
        .replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.') \
        .replace('..', '.').replace('..', '.').replace('..', '.').replace('1. ', '').replace('. 2. ', '. ') \
        .replace('. 3. ', '. ').replace('. 4. ', '. ').replace('. 5. ', '. ').replace(' 2. ', '. ') \
        .replace(' 3. ', '. ').replace(' 4. ', '. ').replace(' 5. ', '. ') \
        .strip().lower().split('. ')
    sent_cleaner = lambda t: re.sub('[.,?;*!%^&_+():-\[\]{}]', '', t.replace('"', '').replace('/', '')
                                    .replace('\\', '').replace("'", '').strip().lower())
    tokens = [sent_cleaner(sent) for sent in report_cleaner(report) if sent_cleaner(sent) != []]
    if tokens == ['']:
        return ""
    report = ' . '.join(tokens) + ' .'
    return report


# https://github.com/ysmiura/ifcc/blob/6c111dbdfe7ce9d3150a5ad90360584cfd2b8442/clinicgen/text/tokenizer.py#L24
# Improving Factual Completeness and Consistency of Image-to-text Radiology Report Generation.
def ifcc_clean_report(report):
    report = report.lower()
    return ' '.join(wordpunct_tokenize(report))


# https://github.com/marshuang80/gloria/blob/main/gloria/models/gloria_model.py#L165
# GLoRIA: A Multimodal Global-Local Representation Learning Framework for Label-efficient Medical Image Recognition
def gloria_clean_report_chexpert(report):
    from nltk.tokenize import RegexpTokenizer
    t = report
    # use space instead of newline
    t = t.replace("\n", " ")

    # split sentences
    splitter = re.compile("[0-9]+\.")
    captions = splitter.split(t)
    captions = [point.split(".") for point in captions]
    captions = [sent for point in captions for sent in point]

    all_sents = []

    for t in captions:
        t = t.replace("\ufffd\ufffd", " ")
        tokenizer = RegexpTokenizer(r"\w+")
        tokens = tokenizer.tokenize(t.lower())

        if len(tokens) <= 1:
            continue

        included_tokens = []
        for t in tokens:
            t = t.encode("ascii", "ignore").decode("ascii")
            if len(t) > 0:
                included_tokens.append(t)
        all_sents.append(" ".join(included_tokens))

    t = " ".join(all_sents)
    return t


# https://github.com/google-research/google-research/blob/master/rouge/tokenize.py
def rouge(text, use_stemmer=False):
    """Tokenize input text into a list of tokens.
    This approach aims to replicate the approach taken by Chin-Yew Lin in
    the original ROUGE implementation.
    Args:
      text: A text blob to tokenize.
      use_stemmer: An optional stemmer.
    Returns:
      A list of string tokens extracted from input text.
    """

    # Pre-compile regexes that are use often
    NON_ALPHANUM_PATTERN = r"[^a-z0-9]+"
    NON_ALPHANUM_RE = re.compile(NON_ALPHANUM_PATTERN)
    SPACES_PATTERN = r"\s+"
    SPACES_RE = re.compile(SPACES_PATTERN)
    VALID_TOKEN_PATTERN = r"^[a-z0-9]+$"
    VALID_TOKEN_RE = re.compile(VALID_TOKEN_PATTERN)

    # Convert everything to lowercase.
    text = text.lower()
    # Replace any non-alpha-numeric characters with spaces.
    text = NON_ALPHANUM_RE.sub(" ", six.ensure_str(text))

    tokens = SPACES_RE.split(text)
    if use_stemmer:
        stemmer = porter.PorterStemmer() if use_stemmer else None
        # Only stem words more than 3 characters long.
        tokens = [six.ensure_str(stemmer.stem(x)) if len(x) > 3 else x
                  for x in tokens]

    # One final check to drop any empty or invalid tokens.
    tokens = [x for x in tokens if VALID_TOKEN_RE.match(x)]

    return tokens



import torch
import torch.nn as nn


class Classifier(nn.Module):
    def __init__(self, input_size, num_classes, dropout=0., **kwargs):
        super(Classifier, self).__init__()
        self.classifier = nn.Sequential(
            nn.Linear(in_features=input_size, out_features=num_classes)
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, input: torch.Tensor):
        output = self.dropout(input)
        output = self.classifier(output)
        return output



import numpy as np

def readable_size(n):
    """Return a readable size string."""
    sizes = ['K', 'M', 'G']
    fmt = ''
    size = n
    for i, s in enumerate(sizes):
        nn = n / (1000 ** (i + 1))
        if nn >= 1:
            size = nn
            fmt = sizes[i]
        else:
            break
    return '%.2f%s' % (size, fmt)


def get_n_params(module):
    n_param_learnable = 0
    n_param_frozen = 0
    for param in module.parameters():
        if not param.data.size():
            continue
        if param.requires_grad:
            n_param_learnable += np.cumprod(param.data.size())[-1]
        else:
            n_param_frozen += np.cumprod(param.data.size())[-1]
    n_param_all = n_param_learnable + n_param_frozen
    return "# parameters: {} ({} learnable)".format(
        readable_size(n_param_all), readable_size(n_param_learnable))




from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import torch
from torch.nn import functional as F
from transformers.generation_utils import GreedySearchOutput, SampleOutput, BeamSearchOutput, BeamSampleOutput, logging, \
    ModelOutput, BeamSearchEncoderDecoderOutput, BeamSearchDecoderOnlyOutput

from transformers.generation_beam_search import BeamScorer, BeamHypotheses, UserDict

from transformers.generation_logits_process import (
    LogitsProcessorList,
)
import copy

logger = logging.get_logger(__name__)


def beam_search(
        self,
        encdecs,
        input_ids: torch.LongTensor,
        beam_scorer: BeamScorer,
        logits_processor: Optional[LogitsProcessorList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        model_kwargs_list=None,
        **model_kwargs,
) -> Union[BeamSearchOutput, torch.LongTensor]:
    # !!!! JB !!!
    # **model_kwargs has lost keys from the function signature
    # need to remove the keys from the dict in model_kwargs_list to match current state of **model_kwargs
    ref_keys = model_kwargs.keys()
    for m in model_kwargs_list:
        to_del_k = [k for k in m.keys() if k not in ref_keys]
        for k in to_del_k:
            del m[k]


    # init values
    logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
    max_length = max_length if max_length is not None else self.config.max_length
    pad_token_id = pad_token_id if pad_token_id is not None else self.config.pad_token_id
    eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id
    output_scores = output_scores if output_scores is not None else self.config.output_scores
    output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate is not None else self.config.return_dict_in_generate
    )

    batch_size = len(beam_scorer._beam_hyps)
    num_beams = beam_scorer.num_beams

    batch_beam_size, cur_len = input_ids.shape

    assert (
            num_beams * batch_size == batch_beam_size
    ), "Batch dimension of `input_ids` should be {num_beams * batch_size}, but is {batch_beam_size}."

    beam_scores = torch.zeros((batch_size, num_beams), dtype=torch.float, device=input_ids.device)
    beam_scores[:, 1:] = -1e9
    beam_scores = beam_scores.view((batch_size * num_beams,))

    while cur_len < max_length:
        models_inputs = [encdec.prepare_inputs_for_generation(input_ids, **m)
                         for encdec, m in zip(encdecs, model_kwargs_list)]

        outputs = [encdec(**model_inputs,
                          return_dict=True,
                          output_attentions=output_attentions,
                          output_hidden_states=output_hidden_states)
                   for encdec, model_inputs in zip(encdecs, models_inputs)]

        next_token_logits = sum(o.logits[:, -1, :] for o in outputs)

        # adjust tokens for Bart, *e.g.*
        next_token_logits = self.adjust_logits_during_generation(
            next_token_logits, cur_len=cur_len, max_length=max_length
        )

        next_token_scores = F.log_softmax(next_token_logits, dim=-1)  # (batch_size * num_beams, vocab_size)
        next_token_scores = logits_processor(input_ids, next_token_scores)
        next_token_scores = next_token_scores + beam_scores[:, None].expand_as(next_token_scores)

        # reshape for beam search
        vocab_size = next_token_scores.shape[-1]
        next_token_scores = next_token_scores.view(batch_size, num_beams * vocab_size)

        next_token_scores, next_tokens = torch.topk(
            next_token_scores, 2 * num_beams, dim=1, largest=True, sorted=True
        )

        next_indices = next_tokens // vocab_size
        next_tokens = next_tokens % vocab_size

        # stateless
        beam_outputs = beam_scorer.process(
            input_ids,
            next_token_scores,
            next_tokens,
            next_indices,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
        )
        beam_scores = beam_outputs["next_beam_scores"]
        beam_next_tokens = beam_outputs["next_beam_tokens"]
        beam_idx = beam_outputs["next_beam_indices"]

        input_ids = torch.cat([input_ids[beam_idx, :], beam_next_tokens.unsqueeze(-1)], dim=-1)

        cur_len = cur_len + 1

        model_kwargs_list = [
            encdec._update_model_kwargs_for_generation(o, m, is_encoder_decoder=self.config.is_encoder_decoder)
            for encdec, o, m in zip(encdecs, outputs, model_kwargs_list)]

        # model_kwargs = self._update_model_kwargs_for_generation(
        #     outputs, model_kwargs, is_encoder_decoder=self.config.is_encoder_decoder
        # )
        #
        for m in model_kwargs_list:
            if m["past"] is not None:
                m["past"] = self._reorder_cache(m["past"], beam_idx)

        if beam_scorer.is_done:
            break

    sequence_outputs = beam_scorer.finalize(
        input_ids, beam_scores, next_tokens, next_indices, pad_token_id=pad_token_id, eos_token_id=eos_token_id
    )

    return sequence_outputs["sequences"]


def generate(
        self,
        encdecs,
        input_ids: Optional[torch.LongTensor] = None,
        max_length: Optional[int] = None,
        min_length: Optional[int] = None,
        do_sample: Optional[bool] = None,
        early_stopping: Optional[bool] = None,
        num_beams: Optional[int] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        bad_words_ids: Optional[Iterable[int]] = None,
        bos_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        length_penalty: Optional[float] = None,
        no_repeat_ngram_size: Optional[int] = None,
        encoder_no_repeat_ngram_size: Optional[int] = None,
        num_return_sequences: Optional[int] = None,
        decoder_start_token_id: Optional[int] = None,
        use_cache: Optional[bool] = None,
        num_beam_groups: Optional[int] = None,
        diversity_penalty: Optional[float] = None,
        prefix_allowed_tokens_fn: Optional[Callable[[int, torch.Tensor], List[int]]] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        **model_kwargs,
) -> Union[GreedySearchOutput, SampleOutput, BeamSearchOutput, BeamSampleOutput, torch.LongTensor]:
    # set init values

    num_beams = num_beams if num_beams is not None else self.config.num_beams
    num_beam_groups = num_beam_groups if num_beam_groups is not None else self.config.num_beam_groups
    max_length = max_length if max_length is not None else self.config.max_length
    do_sample = do_sample if do_sample is not None else self.config.do_sample
    num_return_sequences = (
        num_return_sequences if num_return_sequences is not None else self.config.num_return_sequences
    )

    pad_token_id = pad_token_id if pad_token_id is not None else self.config.pad_token_id
    bos_token_id = bos_token_id if bos_token_id is not None else self.config.bos_token_id
    eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id

    output_scores = output_scores if output_scores is not None else self.config.output_scores
    output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate is not None else self.config.return_dict_in_generate
    )

    model_kwargs["output_attentions"] = output_attentions
    model_kwargs["output_hidden_states"] = output_hidden_states

    if model_kwargs.get("attention_mask", None) is None:
        # init `attention_mask` depending on `pad_token_id`
        model_kwargs["attention_mask"] = self._prepare_attention_mask_for_generation(
            input_ids, pad_token_id, eos_token_id
        )

    # special case if pad_token_id is not defined
    if pad_token_id is None and eos_token_id is not None:
        logger.warning(f"Setting `pad_token_id` to `eos_token_id`:{eos_token_id} for open-end generation.")
        pad_token_id = eos_token_id

    # Storing encoder_input_ids for logits_processor that could use them
    encoder_input_ids = input_ids if self.config.is_encoder_decoder else None

    if self.config.is_encoder_decoder:
        # add encoder_outputs to model_kwargs
        model_kwargs = [encdec._prepare_encoder_decoder_kwargs_for_generation(input_ids, copy.deepcopy(model_kwargs))
                        for encdec in encdecs]

        # set input_ids as decoder_input_ids
        if "decoder_input_ids" in model_kwargs[0]:
            input_ids = model_kwargs[0].pop("decoder_input_ids")
        else:
            input_ids = self._prepare_decoder_input_ids_for_generation(
                input_ids, decoder_start_token_id=decoder_start_token_id, bos_token_id=bos_token_id
            )

        if "encoder_outputs" not in model_kwargs[0] or not isinstance(model_kwargs[0]["encoder_outputs"], ModelOutput):
            raise ValueError("Make sure that `model_kwargs` include `encoder_outputs` of type `ModelOutput`.")

    if input_ids.shape[-1] >= max_length:
        input_ids_string = "decoder_input_ids" if self.config.is_encoder_decoder else "input_ids"
        logger.warning(
            f"Input length of {input_ids_string} is {input_ids.shape[-1]}, but ``max_length`` is set to {max_length}."
            "This can lead to unexpected behavior. You should consider increasing ``config.max_length`` or ``max_length``."
        )

    # determine generation mode
    is_beam_gen_mode = (num_beams > 1) and (num_beam_groups == 1) and do_sample is False

    # set model_kwargs
    for m in model_kwargs:
        m["use_cache"] = use_cache

    # get distribution pre_processing samplers
    logits_processor = self._get_logits_processor(
        repetition_penalty=repetition_penalty,
        no_repeat_ngram_size=no_repeat_ngram_size,
        encoder_no_repeat_ngram_size=encoder_no_repeat_ngram_size,
        encoder_input_ids=encoder_input_ids,
        bad_words_ids=bad_words_ids,
        min_length=min_length,
        max_length=max_length,
        eos_token_id=eos_token_id,
        prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
        num_beams=num_beams,
        num_beam_groups=num_beam_groups,
        diversity_penalty=diversity_penalty,
        remove_invalid_values=None,
        forced_bos_token_id=None,
        forced_eos_token_id=None,
    )

    if is_beam_gen_mode:
        batch_size = input_ids.shape[0]

        length_penalty = length_penalty if length_penalty is not None else self.config.length_penalty
        early_stopping = early_stopping if early_stopping is not None else self.config.early_stopping

        if num_return_sequences > num_beams:
            raise ValueError("`num_return_sequences` has to be smaller or equal to `num_beams`.")
        beam_scorer = BeamSearchScorer(
            batch_size=batch_size,
            max_length=max_length,
            num_beams=num_beams,
            device=self.device,
            length_penalty=length_penalty,
            do_early_stopping=early_stopping,
            num_beam_hyps_to_keep=num_return_sequences,
        )
        # interleave with `num_beams`
        old_inputs = copy.deepcopy(input_ids)
        old_model_kwargs = copy.deepcopy(model_kwargs)

        # need input_ids only once
        input_ids, _ = self._expand_inputs_for_generation(
            old_inputs, expand_size=num_beams, is_encoder_decoder=self.config.is_encoder_decoder, **model_kwargs[0]
        )

        # need model_kwargs updated for each model
        model_kwargs = [e._expand_inputs_for_generation(old_inputs,
                                                        expand_size=num_beams,
                                                        is_encoder_decoder=e.config.is_encoder_decoder, **m)[1]
                        for m, e in zip(old_model_kwargs, encdecs)]

        return beam_search(
            self,
            encdecs,
            input_ids,
            beam_scorer,
            logits_processor=logits_processor,
            max_length=max_length,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
            output_scores=output_scores,
            return_dict_in_generate=return_dict_in_generate,
            model_kwargs_list=model_kwargs,
            **model_kwargs[0],
        )
    else:
        raise NotImplementedError


class BeamSearchScorer(BeamScorer):
    def __init__(
            self,
            batch_size: int,
            max_length: int,
            num_beams: int,
            device: torch.device,
            length_penalty: Optional[float] = 1.0,
            do_early_stopping: Optional[bool] = False,
            num_beam_hyps_to_keep: Optional[int] = 1,
            num_beam_groups: Optional[int] = 1,
    ):
        self.max_length = max_length
        self.num_beams = num_beams
        self.device = device
        self.length_penalty = length_penalty
        self.do_early_stopping = do_early_stopping
        self.num_beam_hyps_to_keep = num_beam_hyps_to_keep
        self.num_beam_groups = num_beam_groups
        self.group_size = self.num_beams // self.num_beam_groups

        self._is_init = False
        self._beam_hyps = [
            BeamHypotheses(
                num_beams=self.num_beams,
                max_length=self.max_length,
                length_penalty=self.length_penalty,
                early_stopping=self.do_early_stopping,
            )
            for _ in range(batch_size)
        ]
        self._done = torch.tensor([False for _ in range(batch_size)], dtype=torch.bool, device=self.device)

        if not isinstance(num_beams, int) or num_beams <= 1:
            raise ValueError(
                f"`num_beams` has to be an integer strictly greater than 1, but is {num_beams}. For `num_beams` == 1, one should make use of `greedy_search` instead."
            )

        if not isinstance(num_beam_groups, int) or (num_beam_groups > num_beams) or (num_beams % num_beam_groups != 0):
            raise ValueError(
                f"`num_beam_groups` has to be an integer smaller or equal than `num_beams` and `num_beams` "
                f"has to be divisible by `num_beam_groups`, but is {num_beam_groups} with `num_beams` being {num_beams}."
            )

    @property
    def is_done(self) -> bool:
        return self._done.all()

    def process(
            self,
            input_ids: torch.LongTensor,
            next_scores: torch.FloatTensor,
            next_tokens: torch.LongTensor,
            next_indices: torch.LongTensor,
            pad_token_id: Optional[int] = None,
            eos_token_id: Optional[int] = None,
    ) -> Tuple[torch.Tensor]:
        cur_len = input_ids.shape[-1]
        batch_size = len(self._beam_hyps)
        assert batch_size == (input_ids.shape[0] // self.group_size)

        device = input_ids.device
        next_beam_scores = torch.zeros((batch_size, self.group_size), dtype=next_scores.dtype, device=device)
        next_beam_tokens = torch.zeros((batch_size, self.group_size), dtype=next_tokens.dtype, device=device)
        next_beam_indices = torch.zeros((batch_size, self.group_size), dtype=next_indices.dtype, device=device)

        for batch_idx, beam_hyp in enumerate(self._beam_hyps):
            if self._done[batch_idx]:
                assert (
                        len(beam_hyp) >= self.num_beams
                ), "Batch can only be done if at least {} beams have been generated".format(self.num_beams)
                assert (
                        eos_token_id is not None and pad_token_id is not None
                ), "generated beams >= num_beams -> eos_token_id and pad_token have to be defined"
                # pad the batch
                next_beam_scores[batch_idx, :] = 0
                next_beam_tokens[batch_idx, :] = pad_token_id
                next_beam_indices[batch_idx, :] = 0
                continue

            # next tokens for this sentence
            beam_idx = 0
            for beam_token_rank, (next_token, next_score, next_index) in enumerate(
                    zip(next_tokens[batch_idx], next_scores[batch_idx], next_indices[batch_idx])
            ):
                batch_beam_idx = batch_idx * self.group_size + next_index
                # add to generated hypotheses if end of sentence
                if (eos_token_id is not None) and (next_token.item() == eos_token_id):
                    # if beam_token does not belong to top num_beams tokens, it should not be added
                    is_beam_token_worse_than_top_num_beams = beam_token_rank >= self.group_size
                    if is_beam_token_worse_than_top_num_beams:
                        continue
                    beam_hyp.add(
                        input_ids[batch_beam_idx].clone(),
                        next_score.item(),
                    )
                else:
                    # add next predicted token since it is not eos_token
                    next_beam_scores[batch_idx, beam_idx] = next_score
                    next_beam_tokens[batch_idx, beam_idx] = next_token
                    next_beam_indices[batch_idx, beam_idx] = batch_beam_idx
                    beam_idx += 1

                # once the beam for next step is full, don't add more tokens to it.
                if beam_idx == self.group_size:
                    break

            if beam_idx < self.group_size:
                raise ValueError(
                    f"At most {self.group_size} tokens in {next_tokens[batch_idx]} can be equal to `eos_token_id: {eos_token_id}`. Make sure {next_tokens[batch_idx]} are corrected."
                )

            # Check if we are done so that we can save a pad step if all(done)
            self._done[batch_idx] = self._done[batch_idx] or beam_hyp.is_done(
                next_scores[batch_idx].max().item(), cur_len
            )

        return UserDict(
            {
                "next_beam_scores": next_beam_scores.view(-1),
                "next_beam_tokens": next_beam_tokens.view(-1),
                "next_beam_indices": next_beam_indices.view(-1),
            }
        )

    def finalize(
            self,
            input_ids: torch.LongTensor,
            final_beam_scores: torch.FloatTensor,
            final_beam_tokens: torch.LongTensor,
            final_beam_indices: torch.LongTensor,
            pad_token_id: Optional[int] = None,
            eos_token_id: Optional[int] = None,
    ) -> Tuple[torch.LongTensor]:
        batch_size = len(self._beam_hyps)

        # finalize all open beam hypotheses and add to generated hypotheses
        for batch_idx, beam_hyp in enumerate(self._beam_hyps):
            if self._done[batch_idx]:
                continue

            # all open beam hypotheses are added to the beam hypothesis
            # beam hypothesis class automatically keeps the best beams
            for beam_id in range(self.num_beams):
                batch_beam_idx = batch_idx * self.num_beams + beam_id
                final_score = final_beam_scores[batch_beam_idx].item()
                final_tokens = input_ids[batch_beam_idx]
                beam_hyp.add(final_tokens, final_score)

        # select the best hypotheses
        sent_lengths = input_ids.new(batch_size * self.num_beam_hyps_to_keep)
        best = []
        best_scores = torch.zeros(batch_size * self.num_beam_hyps_to_keep, device=self.device, dtype=torch.float32)

        # retrieve best hypotheses
        for i, beam_hyp in enumerate(self._beam_hyps):
            sorted_hyps = sorted(beam_hyp.beams, key=lambda x: x[0])
            for j in range(self.num_beam_hyps_to_keep):
                best_hyp_tuple = sorted_hyps.pop()
                best_score = best_hyp_tuple[0]
                best_hyp = best_hyp_tuple[1]
                sent_lengths[self.num_beam_hyps_to_keep * i + j] = len(best_hyp)

                # append to lists
                best.append(best_hyp)
                best_scores[i * self.num_beam_hyps_to_keep + j] = best_score

        # prepare for adding eos
        sent_max_len = min(sent_lengths.max().item() + 1, self.max_length)
        decoded: torch.LongTensor = input_ids.new(batch_size * self.num_beam_hyps_to_keep, sent_max_len)
        # shorter batches are padded if needed
        if sent_lengths.min().item() != sent_lengths.max().item():
            assert pad_token_id is not None, "`pad_token_id` has to be defined"
            decoded.fill_(pad_token_id)

        # fill with hypotheses and eos_token_id if the latter fits in
        for i, hypo in enumerate(best):
            decoded[i, : sent_lengths[i]] = hypo
            if sent_lengths[i] < self.max_length:
                decoded[i, sent_lengths[i]] = eos_token_id
        return UserDict(
            {
                "sequences": decoded,
                "sequence_scores": best_scores,
            }
        )
